/*
   Init for the Official LFS LiveCD
   Written by Jeremy Huntwork, 2005-09-16

   This code sets up a directory structure for the CD in the initramfs,
   finds and mounts the LFS LiveCD, mounts the root file system via the
   squashfs file, overlays it all with unionfs and finally passes
   control over to sysvinit to finish the boot process.
*/

#include "init.h"


/* Devices to check for the LFS CD, could scan some stuff in /proc in a later version */
/* Some devices have been deleted */
const char *devices[] = 
{
	"/dev/hda", 
	NULL
};

const char *lfscd;

int mountlfscd(void);

int losetup(char * loop, char * file, int flags)
{
	struct loop_info loopinfo;
	int fd, ffd;
	
	memset(&loopinfo, 0, sizeof(loopinfo));

	ffd = open(file, flags);
	if (ffd<0) {
		printf("Failed to open the %s file: %s\n", file, strerror(errno));
		return(0);
	}
	
	fd = open(loop, flags);
	if (fd<0) {
		printf("Failed to open the loop device: %s\n", strerror(errno));
		return(0);
	}

	snprintf(loopinfo.lo_name, LO_NAME_SIZE, "%s", file);

	loopinfo.lo_offset = 0;
	loopinfo.lo_encrypt_key_size = 0;
	loopinfo.lo_encrypt_type = LO_CRYPT_NONE;

	if(ioctl(fd, LOOP_SET_FD, ffd) < 0) {
		printf("Failed to set up device: %s\n", strerror(errno));
		return(0);
	}
	close(ffd);

        if(ioctl(fd, LOOP_SET_STATUS, &loopinfo) < 0) {
                printf("Failed to set up device: %s\n", strerror(errno));
		(void) ioctl(fd, LOOP_CLR_FD, 0);
		close(fd);
                return(0);
        }
        close(fd);
	return(1);
}

int main(int argc, char * argv[], char * envp[])
{
	int i, overhead;
	int fd;
	struct dm_task * dmt;
	int devsize;
	
	printf("Initramfs activated\n");

	mkdir("/proc", 0755);
	mount("proc", "/proc", "proc", 0, 0);
	
	mkdir(TMPFS, 0755);
	if (mount("tmpfs", TMPFS, "tmpfs", 0, "size=90%") < 0) {
		printf("Failed to mount tmpfs: %s\n", strerror(errno));
		return (0);
	}
	
	mkdir(CDROM_MOUNT, 0755);

	/* Try to losetup iso file */
	/* use loop9 */
	printf("try to losetup iso file");
	mount("/dev/hda11", "/MYLIN", "ext3", 0, 0);
	losetup("/dev/loop9", "/MYLIN/lfslivecd-x86-6.2-3.iso", O_RDONLY);
	printf("iso file losetuped to loop9");

	/* Now, attempt to attach the root file to /dev/loop0 */
	printf("Setting up the loopback devices...\n");
	losetup("/dev/loop0", ROOT_FILE, O_RDONLY);
	
	/* Create a sparse file for the second loop */
	
	fd = open("/dev/loop0", O_RDONLY);
	ioctl(fd, BLKGETSIZE, &devsize);
	close(fd);
	
	overhead = 0x1000 + devsize * 2;
	
	fd = open(OVERLAY, O_CREAT | O_WRONLY, 0600);
	ftruncate(fd, devsize * 0x200ULL + overhead);
	close(fd);
	
	losetup("/dev/loop1", OVERLAY, O_RDWR);
	
	/* Set up device-mapper */
        dmt = dm_task_create(DM_DEVICE_CREATE);
        dm_task_set_name(dmt, "lfs-cd");
        dm_task_set_major(dmt, 254);
        dm_task_set_minor(dmt, 0);
        dm_task_add_target(dmt, 0, devsize,
	    "snapshot", "/dev/loop0 /dev/loop1 p 8");
        dm_task_run(dmt);
        dm_task_destroy(dmt);
							
	printf("Mounting root filesystem...\n");
	mkdir(ROOT, 0755);
	i = mount("/dev/mapper/lfs-cd", ROOT, "ext2", 0, 0);
	if (i<0) {
		printf("Failed to mount root fs: %s\n", strerror(errno));
		return(0);
	}

	/* Move the tmpfs to /dev/shm in the root fs */

	mount("/.tmpfs", ROOT "/dev/shm", NULL, MS_MOVE, NULL);

	/* Create a symlink for the CD drive to /dev/lfs-cd */

	symlink(lfscd, ROOT "/dev/lfs-cd");


	/* Remove the "/init" binary to free some RAM */
	unlink ("/init");
	umount2("/proc", 0);
	
	/* Chroot into the root fs */

	chdir(ROOT);
	mount(".", "/", NULL, MS_MOVE, NULL);

	if ( chroot(".") || chdir("/") )
		return(0);
	
	/* We're done! Pass control to sysvinit. */

	printf("Starting init...\n");
	
	/* FIXME: file descriptors still point to initramfs */
	execve("/sbin/init", argv, envp);
	printf("Failed to start init: %s :(\n", strerror(errno));

	return(0);
}

/* This will mount the LFS LiveCD and return 1 */
/* it has been changed a lot */
int mountlfscd(void)
{
	struct iso_primary_descriptor pd;
	int fd, status;
	char buf[ISO_BLOCK_SIZE];
	fd = open("/dev/loop9", O_RDONLY | O_NONBLOCK);
	status = ioctl(fd, CDROM_DRIVE_STATUS, 0);
	lseek(fd, ISO_PD_BLOCK*ISO_BLOCK_SIZE, SEEK_SET);
	read(fd, buf, ISO_BLOCK_SIZE);
	memcpy( &pd, buf, sizeof(pd) );
	status = mount("/dev/loop9", CDROM_MOUNT, CDROM_FSTYPE, MS_RDONLY, NULL);
	memset(buf, 0, 32); /* Clear buffer so we don't have to check read's status */
	read(fd, buf, 10);
	close(fd);
	lfscd = "/dev/loop9";
	return(1);
}


